/*
  SPI interface routines for SD & SDHC & MMC cards

  Jonathan "lonesock" Dummer
  version 0.3.0  2009 July 19

  Using multiblock SPI mode exclusively.

  This is the "SAFE" version...uses
  * 1 instruction per bit writes
  * 2 instructions per bit reads

  For the fsrw project:
  fsrw.sf.net
*/

static int clkfreq = 80000000;
#include <stdio.h>
#include <stdlib.h>
#include <propeller.h>

int sdspi_start( int basepin );
int sdspi_start_explicit( int DO, int CLK, int DI, int CS );
void errorexit( int val );
int send_cmd_slow( int cmd, int val, int crc );
void send_slow( int value, int bits_to_send );
int read_32_slow( void);
int read_slow( void);

// possible card types
#define type_MMC      1
#define type_SD       2
#define type_SDHC     3
  
// Error codes
#define ERR_CARD_NOT_RESET            -1
#define ERR_3v3_NOT_SUPPORTED         -2
#define ERR_OCR_FAILED                -3
#define ERR_CARD_BUSY_TIMEOUT         -1000

// SDHC/SD/MMC command set for SPI
#define CMD0    (0x40+0)       // GO_IDLE_STATE 
#define CMD1    (0x40+1)       // SEND_OP_COND (MMC) 
#define ACMD41  (0xC0+41)      // SEND_OP_COND (SDC) 
#define CMD8    (0x40+8)       // SEND_IF_COND 
#define CMD12   (0x40+12)      // STOP_TRANSMISSION
#define CMD16   (0x40+16)      // SET_BLOCKLEN 
#define CMD55   (0x40+55)      // APP_CMD 
#define CMD58   (0x40+58)      // READ_OCR
#define CMD59   (0x40+59)      // CRC_ON_OFF 

static int start_cnt       = 0;  // cnt at init
static int all_mask        = 0;
static int adrShift        = 9;  // will be 0 for SDHC, 9 for MMC & SD

static int di_pin, do_pin, clk_pin, cs_pin, starttime;
static int di_mask, do_mask, clk_mask, cs_mask;
static int aflag;
static int sdhc = 0;
static int delay_cycles = 0;

void abortexit(int value)
{
  printf("abortexit: %d\n", value);
  exit(value);
}
  
/*
  This is a compatibility wrapper, and requires that the pins be
  both consecutive, and in the order DO CLK DI CS.
*/
int sdspi_start( int basepin )
{
  return sdspi_start_explicit( basepin, basepin+1, basepin+2, basepin+3 );
}

/*
  Do all of the card initialization in SPIN, then hand off the pin
  information to the assembly cog for hot SPI block R/W action!
*/
int sdspi_start_explicit( int DO, int CLK, int DI, int CS )
{
  int card_type, tmp, i, j;
  // wait ~4 milliseconds
  waitcnt( 500 + (clkfreq>>8) + CNT );
  // (start with cog variables, _BEFORE_ loading the cog)
  do_pin  = DO;
  clk_pin = CLK;
  di_pin  = DI;
  cs_pin  = CS;

  if (DO < 32)
  {
    aflag = 1;
    do_mask = 1 << DO;
    di_mask = 1 << DI;
    cs_mask = 1 << CS;
    clk_mask = 1 << CLK;
  }
  else
  {
    aflag = 0;
    do_mask = 1 << (DO - 32);
    di_mask = 1 << (DI - 32);
    cs_mask = 1 << (CS - 32);
    clk_mask = 1 << (CLK - 32);
  }
  adrShift = 9; // block = 512 * index, and 512 = 1<<9
  // pass the output pin mask via the command register
  all_mask = cs_mask | clk_mask | di_mask;
  DIRB |= all_mask  ;
  // get the card in a ready state: set DI and CS high, send => 74 clocks
  OUTB |= all_mask;
  for (j = 0; j < 4096; j++)
  {
    OUTB |= clk_mask;
    OUTB &= ~clk_mask;
  }
  // time-hack
  start_cnt = CNT;
  // reset the card
  tmp = 0;
  for (i = 0; i <= 9; i++)
  {
    if (tmp != 1)
    {
      tmp = send_cmd_slow( CMD0, 0, 0x95 );
      if (tmp & 4)
      {
        // the card said CMD0 ("go idle") was invalid, so we're possibly stuck in read or write mode
        if (i & 1)
        {
          // exit multiblock read mode
          for (j = 0; j < 4; j++)
            read_32_slow();      // these extra clocks are required for some MMC cards
          send_slow( 0xFD, 8 );  // stop token
          read_32_slow();
          while (read_slow() != 0xFF);
        }
        else
        {
          // exit multiblock read mode
          send_cmd_slow( CMD12, 0, 0x61 );
        }
      }
    }
  }
  if (tmp != 1)
  {
    // the reset command failed!
    errorexit( ERR_CARD_NOT_RESET );
  }
  // Is this a SD type 2 card?
  if (send_cmd_slow( CMD8, 0x1AA, 0x87 ) == 1)
  {
    // Type2 SD, check to see if it's a SDHC card
    tmp = read_32_slow();
  // check the supported voltage
    if ((tmp & 0x1FF) != 0x1AA)
    {
      errorexit( ERR_3v3_NOT_SUPPORTED );
    }
    // try to initialize the type 2 card with the High Capacity bit
    while (send_cmd_slow( ACMD41, 1 << 30, 0x77 ));
    // the card is initialized, let's read back the High Capacity bit
    if (send_cmd_slow( CMD58, 0, 0xFD ) != 0)
    {
      errorexit( ERR_OCR_FAILED );
    }
    // get back the data
    tmp = read_32_slow();
    // check the bit
    if (tmp & (1 << 30))
    {
      card_type = type_SDHC;
      adrShift = 0;
      sdhc = 1;
    }
    else
      card_type = type_SD;
  }
  else
  {
    // Either a type 1 SD card, or it's MMC, try SD 1st
    if (send_cmd_slow( ACMD41, 0, 0xE5 ) < 2)
    {
      // this is a type 1 SD card (1 means busy, 0 means done initializing)
      card_type = type_SD;
      while (send_cmd_slow( ACMD41, 0, 0xE5 ));
    }
    else
    {
      // mark that it's MMC, and try to initialize
      card_type = type_MMC;
      while (send_cmd_slow( CMD1, 0, 0xF9 ));
    }
    // some SD or MMC cards may have the wrong block size, set it here
    send_cmd_slow( CMD16, 512, 0x15 );
  }
  // card is mounted, make sure the CRC is turned off
  send_cmd_slow( CMD59, 0, 0x91 );
#if 0
  //  check the status
  //send_cmd_slow( CMD13, 0, 0x0D )    
  // done with the SPI bus for now
  OUTB |= cs_mask;
  DIRB &= ~all_mask;
#endif
  // the return variable is card_type   
  return card_type;
}

/*
  Send down a command and return the reply.
  Note: slow is an understatement!
  Note: this uses the assembly DAT variables for pin IDs,
  which means that if you run this multiple times (say for
  multiple SD cards), these values will change for each one.
  But this is OK as all of these functions will be called
  during the initialization only, before the PASM engine is
  running.
*/
int send_cmd_slow( int cmd, int val, int crc )
{
  int reply, time_stamp;
  // if this is an application specific command, handle it
  if (cmd & 0x80)
  {
    // ACMD<n> is the command sequense of CMD55-CMD<n>
      cmd &= 0x7F;
      reply = send_cmd_slow( CMD55, 0, 0x65 );
      if (reply > 1)
        return reply;
  }
  // the CS line needs to go low during this operation
  OUTB |= cs_mask;
  OUTB &= ~cs_mask;
  // give the card a few cocks to finish whatever it was doing
  read_32_slow();
  // send the command byte
  send_slow( cmd, 8 );
  // send the value long
  send_slow( val, 32 );
  // send the CRC byte
  send_slow( crc, 8 );
  // is this a CMD12?, if so, stuff byte
  if (cmd == CMD12)
    read_slow();
  // read back the response (spec declares 1-8 reads max for SD, MMC is 0-8)
  time_stamp = 9;
  do
  {
    reply = read_slow();
  }
  while (( reply & 0x80 ) && ( time_stamp-- ));
  // done, and 'reply' is already pre-loaded
  return reply;
}

void send_slow( int value, int bits_to_send )
{
  int i;

  value <<= 32 - bits_to_send;
  for (i = 0; i < bits_to_send; i++)
  {
    OUTB &= ~clk_mask;
    if (value & 0x80000000)
      OUTB |= di_mask;
    else
      OUTB &= ~di_mask;
    value <<= 1;
    OUTB |= clk_mask;
  }
}

int read_32_slow( void)
{
  int i;
  int r = 0;
  for (i = 0; i < 4; i++)
  {
    r <<= 8;
    r |= read_slow();
  }
  return r;
}
  
/*
  Read back 8 bits from the card
*/
int read_slow( void)
{
  int i;
  int r = 0;
  // we need the DI line high so a read can occur
  OUTB |= di_mask;
  // get 8 bits (remember, r is initialized to 0 by SPIN)
  for (i = 0; i < 8; i++)
  {
    OUTB &= ~clk_mask;
    OUTB |= clk_mask;
    r <<= 1;
    if (INB & do_mask) r |= 1;
  }
  // error check
  if( (CNT - start_cnt) > (clkfreq << 2) )
    errorexit( ERR_CARD_BUSY_TIMEOUT );
  return r;
}
//  sdspi:  SPI interface to a Secure Digital card.
//
//  Copyright 2008   Radical Eye Software
//
//  See end of file for terms of use.
//
//  This version is in Spin so it is very slow (3Kb/sec).
//  A version in assembly is about 100x faster.
//
//  You probably never want to call this; you want to use fswr
//  instead (which calls this); this is only the lowest layer.
//
//  Assumes SD card is interfaced using four consecutive Propeller
//  pins, as follows (assuming the base pin is pin 0):
//
//  The 150 ohm resistors are current limiters and are only
//  needed if you don't trust your code (and don't want an SD
//  driven signal to conflict with a Propeller driven signal).
//  A value of 150 should be okay, unless you've got some
//  unusually high capacitance on the line.  The 20k resistors
//  are pullups, and should be there on all six lines (even
//  the ones we don't drive).
//
//  This code is not general-purpose SPI code; it's very specific
//  to reading SD cards, although it can be used as an example.
//
//  The code does not use CRC at the moment (this is the default).
//  With some additional effort we can probe the card to see if it
//  supports CRC, and if so, turn it on.   
//
//  All operations are guarded by a watchdog timer, just in case
//  no card is plugged in or something else is wrong.  If an
//  operation does not complete in one second it is aborted.
//

#define CYCLES_PER_MICROSECOND 80

void sdspi_delay(void)
{
    if (delay_cycles)
    {
        waitcnt(CNT+delay_cycles);
    }
}

void errorexit(int val)
{
    printf("errorexit: %d\n", val);
    exit(1);
}

int sdspi_getcnt(void)
{
    return CNT;
}

void release(void)
{
    if (aflag)
    {
        OUTA |= clk_mask | di_mask | cs_mask;
        DIRA &= ~(clk_mask | di_mask | cs_mask);
    }
    else
    {
        OUTB |= clk_mask | di_mask | cs_mask;
        DIRB &= ~(clk_mask | di_mask | cs_mask);
    }
}

//
//  Send eight bits, then raise di.
//
void senda(int outv)
{
    int i;

    for (i = 0; i < 8; i++)
    {
        OUTA &= ~clk_mask;
        sdspi_delay();
        if (outv & 0x80)
            OUTA |= di_mask;
        else
            OUTA &= ~di_mask;
        outv <<= 1;
        OUTA |= clk_mask;
        sdspi_delay();
    }

    OUTA |= di_mask;
}

void sendb(int outv)
{
    int i;

    for (i = 0; i < 8; i++)
    {
        OUTB &= ~clk_mask;
        sdspi_delay();
        if (outv & 0x80)
            OUTB |= di_mask;
        else
            OUTB &= ~di_mask;
        outv <<= 1;
        OUTB |= clk_mask;
        sdspi_delay();
    }

    OUTB |= di_mask;
}

void send(int outv)
{
    if (aflag)
        senda(outv);
    else
        sendb(outv);
}

//
//  Did we go over our time limit yet?
//
void checktime(void)
{
    if (sdspi_getcnt() - starttime > 50000000)
        errorexit(-41); // Timeout during read
}

//
//  Read eight bits from the card.
//
int reada(void)
{
    int i, r;

    r = 0;
    for (i = 0; i < 8; i++)
    {
        OUTA &= ~clk_mask;
        sdspi_delay();
        OUTA |= clk_mask;
        sdspi_delay();
        r <<= 1;
        if (INA & do_mask)
            r |= 1;
    }
    return r;
}

int readb(void)
{
    int i, r;

    r = 0;
    for (i = 0; i < 8; i++)
    {
        OUTB &= ~clk_mask;
        sdspi_delay();
        OUTB |= clk_mask;
        sdspi_delay();
        r <<= 1;
        if (INB & do_mask)
            r |= 1;
    }
    return r;
}

int read(void)
{
    int retval;

    if (aflag)
        retval = reada();
    else
        retval = readb();

    return retval;
}

//
//  Read eight bits, and loop until we
//  get something other than $ff.
//
int readresp(void)
{
    int r;

    while (1)
    {
        r = read();
        if (r != 0xff) break;
        checktime();
    }

    return r;
}

    
//
//  Wait until card stops returning busy
//
int busy(void)
{
    int r;

    while (1)
    {
        r = read();
        if (r) break;
        checktime();
    }

    return r;
}

//
//  Send a full command sequence, and get and
//  return the response.  We make sure cs is low,
//  send the required eight clocks, then the
//  command and parameter, and then the CRC for
//  the only command that needs one (the first one).
//  Finally we spin until we get a result.
//
int cmd(int op, int parm)
{
    if (aflag)
        OUTA &= ~cs_mask;
    else
        OUTB &= ~cs_mask;
    read();
    send(0x40+op);
    send(parm >> 24);
    send(parm >> 16);
    send(parm >> 8);
    send(parm << 0);
    if (op == 0)
        send(0x95);
    else
        send(0x87);
    return readresp();
}

//
//  Deselect the card to terminate a command.
//
int endcmd(void)
{
    if (aflag)
        OUTA |= cs_mask;
    else
        OUTB |= cs_mask;
    return 0;
}

//
//  Read a single block.  The "n" passed in is the
//  block number (blocks are 512 bytes); the b passed
//  in is the address of 512 blocks to fill with the
//  data.
//
int readblock(int n, char *b)
{
    int i;

    starttime = sdspi_getcnt();
    if (!sdhc) n <<= 9;
    cmd(17, n);
    readresp();
    for (i = 0; i < 512; i++)
        *b++ = read();
    read();
    read();
    return endcmd();
}

//
//  Read the CSD register.  Passed in is a 16-byte
//  buffer.
//
int getCSD(char *b)
{
    int i;

    starttime = sdspi_getcnt();
    cmd(9, 0);
    readresp();
    for (i = 0; i < 16; i++)
        *b++ = read();
    read();
    read();
    return endcmd();
}

//
//  Write a single block.  Mirrors the read above.
//
int writeblock(int n, char *b)
{
    int i;

    starttime = sdspi_getcnt();
    if (!sdhc) n <<= 9;
    cmd(24, n);
    send(0xfe);
    for (i = 0; i < 512; i++)
        send(*b++);
    read();
    read();
    if ((readresp() & 0x1f) != 5)
        errorexit(-42);
    busy();
    return endcmd();
}

//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files
// (the "Software"), to deal in the Software without restriction,
// including without limitation the rights to use, copy, modify, merge,
// publish, distribute, sublicense, and/or sell copies of the Software,
// and to permit persons to whom the Software is furnished to do so,
// subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
